//*********************************************
//  Plan.C
// 
// Ravi Murthy
//*********************************************


#include <assert.h>

#include "Globals.h"
#include "TupleIterator.h"

#ifdef    WITH_PERSISTENCE
#include "Plan.h"
#include "NLExtJoin.h"
#include "GraceJoin.h"

extern Cor_suspensionClass *Cor_suspension;
#endif

// *************************************************************************
// The external join solver is called when get_next_tuple on the 
// ext_join builtin is invoked. This solver must first figure out if
// this is the first time it is called or not. If it is the first time,
// the join tree should be generated and a result tuple returned. If it
// is not the first time, the next result tuple from the join should
// be returned.
Cor_BindEnv *Cor_ExtJoinSolver(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    WITH_PERSISTENCE
    Cor_ExtJoin *root;

    // First check if this is the first time this particular external 
    // join is being called.
    if (it.ipos == 0) {
        root = create_plan(it);
        if (root == NULL) {
            *(Cor_exEnv->err_strm) << "Plan not generated\n";
            it.set_no_match();
            return (it.bindenv);
        }
        it.userData = (void *)root;
        it.ipos = -1;
    } else {
        root = (Cor_ExtJoin *)it.userData;
        it.ipos = -1;
    }


    Cor_Tuple *tuple = root->get_next_tuple(it);

    if (it.no_match()) {
        delete root;
        it.ipos = 0;
    }

#else

    *(Cor_exEnv->err_strm)
        << "The command '" << *Cor_ExtJoinSymbol << "' is unavailable.\n"
        << "Persistence is not compiled into this version of Coral.\n";

#endif   // WITH_PERSISTENCE

    return (it.bindenv);
}

#ifdef    WITH_PERSISTENCE
//************************************************************
// This function creates a default access plan for the given external
// join sequence. In particular, it takes the order in which the relations
// are specified as the default order for the left deep tree that is 
// constructed. At the same time, it also checks if the join can be 
// performed (all valid arities, join attrs of the same type, etc).
// If the join cannot be done, it returns NULL.
// Otherwise, it returns a pointer to a node of type Cor_ExtJoin which is 
// actually the root of the plan tree.

Cor_ExtJoin *create_plan(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

// **************
// NOTE: THE DEFAULT JOIN METHOD IS NOW GRACE JOIN FOR THE FIRST PAIR
// OF RELATIONS AND NESTED LOOP JOINS FOR EVERY OTHER JOIN.
// The arguments are basically a list of relations with their attributes.
// First, check if the mentioned relations are in exodus.
// Try to access both the relations making sure that the arity values 
// are correct.
// Next, check that the join can indeed be done on the specified attributes.
// ie. the types of the attributes to be joined match.
// At the end of this step, we are sure that a join can be done
//****************

    // Check the number of arguments to ext_join()
    int argcount = it.arg_list.count();
    if (argcount < 2)
    {
        *(Cor_exEnv->err_strm) << "Error: insufficient number of arguments\n";
        return(NULL);
    }
    
/*
    // First argument should be the join method.
    char *joinname = it.arg_list[0].SymbolP()->str;
    if(strcmp(joinname, "nl"))
    {
        *(Cor_exEnv->err_strm) << "Error: insufficient number of arguments\n";
        *(Cor_exEnv->err_strm) << "Error: Unknown join method\n";
        return(it.bindenv);
    }

*/
    int i, j;
    Cor_CArg *func;
    Cor_CArg *var1;
    Cor_Relation *joinrel;       // To store outer and inner relations
    Cor_ExtJoin *extjoin = NULL, *prevextjoin = NULL;
    Cor_TupleIterator *previter, *iter;

    int size = it.arg_list.max_vars();
    DBTermType *types = new DBTermType[size];

    // Initialize types
    for (i = 0; i < size; i++)
        types[i] = COR_INVALID_DBTYPE;

    for(i=0;i<it.arg_list.count();i++)
    {
        func = &(it.arg_list[i]);
        // Try to open the relation 
        if ((joinrel = find_relation(func->functor(), func->arity())) == NULL)
        {
            *(Cor_exEnv->err_strm) << "Error: Unable to open relation "
                << func->functor() << endl;
            delete types;
            delete prevextjoin;
            return(NULL);
        }

        // Check if arity is right
        if(joinrel->arity() != func->arity())
        {
            *(Cor_exEnv->err_strm) << "Error: Arity of relation "
                << func->functor() << " is incorrect\n";
            delete types;
            delete prevextjoin;
            return(NULL);
        }

        // Check if the relation is in Exodus
        if (joinrel->r_kind != COR_R_EXODUS)
        {
            *(Cor_exEnv->err_strm) << "Error: Relation "
                    << func->functor()
                    << " is not in exodus\n";
            delete types;
            delete prevextjoin;
            return(NULL);
        }


        // Check if the types of join attributes match.
        for (j=0; j< func->arity(); j++) {
            var1 = &((func->args())[j]);
            if (types[var1->var_num()] !=  COR_INVALID_DBTYPE) {
                if (types[var1->var_num()] !=
                        ((Cor_PersistentRel *)joinrel)->file_desc().schema[j]) {
                    *(Cor_exEnv->err_strm)
                        << "Error: Type mismatch in field " << j
                        << " in relation "
                        << joinrel->name << endl;;
                    delete  types;
                    delete prevextjoin;
                    return NULL;
                }
            } else {
                types[var1->var_num()] =
                    ((Cor_PersistentRel *)joinrel)->file_desc().schema[j];
            }
        }


        // Create new join nodes.
        iter = new Cor_TupleIterator(joinrel, func->args(), it.bindenv);
        assert(iter != NULL);
        if (i == 1)
            extjoin = new Cor_GraceJoin(previter, iter);
        else if (i > 1)
            extjoin = new Cor_NLExtJoin(prevextjoin, iter);

        previter = iter;
        prevextjoin = extjoin;
    }

    delete  types;
    return extjoin;
}

// *************************************************************************
#endif // WITH_PERSISTENCE
