// *************************************************************************
//      ParserStruct.C
// *************************************************************************

#include "ParserStruct.h"

#include "Predicate.h"
#include "Literal.h"
#include "Globals.h"
#include "Collection.h"
#include "Debug.h"

// *************************************************************************
Cor_ParserStruct::Cor_ParserStruct()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    func          = NULL;
    label_counter = 1;

    snrule        = NULL;
    table         = NULL;
    rInfo_arr     = NULL;

    nested_index  = 0;

    cur_rule      = NULL;
    cur_collect   = new Cor_Collection();
    all_rules     = new Cor_CollectStack();

    CurModule     = new Cor_ModuleInfo();

    Predicates      = NULL;
    PredicateTable  = NULL;
    AllPredicates   = NULL;
    Junk_Preds      = NULL;

    make_index_annotations  = NULL;
    agg_sel_annotations     = NULL;
    diskrel_annotations     = NULL;
    asyncrel_annotations    = NULL;
    AllowedAdornList        = NULL;
    DisallowedAdornList     = NULL;
    MultisetAnnotations     = NULL;

    lhs_flag                = 0;
    agg_flag                = 0;

    no_param_args = 0;
    orig_args     = NULL;
}

// *************************************************************************
//  This function should only be called from reset()
void Cor_ParserStruct::init()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    label_counter = 1;

      // Many of these may need to have destructors called on them
      // These should all have been deleted already by clear()

    AllPredicates = NULL;
    Predicates = NULL;

    make_index_annotations = NULL;
    agg_sel_annotations    = NULL;
    AllowedAdornList       = NULL;
    DisallowedAdornList    = NULL;
    MultisetAnnotations    = NULL;
    diskrel_annotations    = NULL;
    asyncrel_annotations   = NULL;

    PredicateTable = NULL;

    CurModule->init();
    func = NULL;

    lhs_flag = 0;
    agg_flag = 0;

    nested_index = 0;
    cur_rule = NULL;

    snrule = NULL;
    table = NULL;
    rInfo_arr = NULL;
}

// *************************************************************************
Cor_ParserStruct::~Cor_ParserStruct()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    clear(1);

    delete orig_args;

    delete all_rules;
    delete cur_collect;

    delete CurModule;

    // Willing to have these leak for now, but don't want to hear
    // about them when running purify
#ifdef    DO_PURE
      // Delete all junk predicates
    Cor_Predicate *last = NULL;

    for (Cor_Predicate *pred = Junk_Preds;
                                pred; pred=pred->allprednext) {
        delete last;
        last = pred;
    }

    delete last;
#endif
}

// *************************************************************************
void Cor_ParserStruct::clear(int serious)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_exEnv->C_exec_mode != COR_DOING_IMPERATIVE)
        delete cur_rule;
    cur_rule = NULL;

    delete PredicateTable;
    PredicateTable = NULL;


      // If someone doesn't want these deleted, they
      // would have set these pointers to NULL.
    make_index_annotations->delete_list();
    agg_sel_annotations->delete_list();
    MultisetAnnotations->delete_list();
    diskrel_annotations->delete_list();
    asyncrel_annotations->delete_list();

      // Magic doesn't have to set these to NULL though
      // (these are only used by magic currently)
    if (! Cor_exEnv->i_am_magic()) {
        AllowedAdornList->delete_list();
        DisallowedAdornList->delete_list();
    }

        // In server mode we can have suspends/resumes to 
        // complicate things
    if (serious && !coralConfig->server_mode) {

        if (snrule) {
            delete snrule;
            snrule = NULL;
            delete table;
            delete rInfo_arr;
        }

          // Delete all predicates
        Cor_Predicate *last = NULL;

        for (Cor_Predicate *pred = AllPredicates;
                                pred; pred=pred->allprednext) {
            delete last;
            last = pred;
        }

        delete last;
        AllPredicates = NULL;
    } else {

    // Willing to have these leak for now, but don't want to hear
    // about them when running purify

#ifdef    DO_PURE
          // Save the predicates to be deleted later.
        Cor_Predicate *last = NULL;
        for (Cor_Predicate *pred = AllPredicates;
                                pred; pred=pred->allprednext) {
            if (last) {
                last->allprednext = Junk_Preds;
                Junk_Preds = last;
            }
            last = pred;
        }

        if (last) {
            last->allprednext = Junk_Preds;
            Junk_Preds = last;
        }
#endif
        
    }

}

// *************************************************************************
void Cor_ParserStruct::reset(int serious)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    clear(serious);
    CurModule->clear();

    init();
}

// *************************************************************************
//  Find the (globally unique) Predicate with the given 'name' and 'arity'.
//  Allocate a new one if none is found.  LookupPredicate() is similar,
//  except that it does NOT allocate a new predicate if it is not found.
// *************************************************************************
Cor_Predicate *Cor_FindPredicate(Cor_ParserStruct *Cor_parserStruct,
        Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Predicate *pred;

    if (Cor_parserStruct->PredicateTable == NULL)
        Cor_parserStruct->PredicateTable = new Cor_SymTable();

    pred = (Cor_Predicate *)
            Cor_parserStruct->PredicateTable->lookup(name);

    if (! pred) {

          // No previous Predicate with that name
        pred = new Cor_Predicate(Cor_parserStruct, name, arity);

        Cor_parserStruct->PredicateTable->insert(name, (Pointer) pred);
        pred->symtabnext = Cor_parserStruct->Predicates;
        Cor_parserStruct->Predicates = pred;
        return pred;
    }

      // Check all existing Predicates with same 'name' for given 'arity'
    for (;; pred = pred->symtabnext) {
        if (pred->arity == arity) {
            return pred;
        } else if ((pred->symtabnext == NULL)
                            || (! pred->symtabnext->name.equals(name))) {
                // No such Predicate. Allocate and link in a new one.
            Cor_Predicate *npred =
                    new Cor_Predicate(Cor_parserStruct, name, arity);

            npred->symtabnext = pred->symtabnext;
            pred->symtabnext = npred;
            return npred;
        }
    }
}

// *************************************************************************
//  Same as FindPredicate, except that if it is not found, a new predicate
//  is NOT created.  Instead NULL is returned.
// *************************************************************************
Cor_Predicate *Cor_LookupPredicate(Cor_ParserStruct *parserStruct,
        Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Predicate *pred;

    if (parserStruct->PredicateTable == NULL)
        parserStruct->PredicateTable = new Cor_SymTable();

    pred = (Cor_Predicate *)
                parserStruct->PredicateTable->lookup(name);

    if (! pred)
        return(NULL);

      // Check all existing Predicates with same 'name' for given 'arity'

    for (;; pred = pred->symtabnext) {
        if ((arity == -1) || (pred->arity == arity))
            return(pred);
        else if ((pred->symtabnext == NULL)
                || (! pred->symtabnext->name.equals(name))) {
              // No such Predicate
            return(NULL);
        }
    }
}

// *************************************************************************
