// *************************************************************************
//    FileDescStruct.C
// *************************************************************************

#ifdef WITH_PERSISTENCE

#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"
#include "Debug.h"


// *************************************************************************
// Constructor takes a schema definition of a relation as parameter.
// The schema definition is in the form of a CArg (that is a functor)

Cor_FileDescStruct::Cor_FileDescStruct(Cor_CArg *arg,
        int& error, int temporary, int by_example)
        : schema(arg->args(), error, by_example)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    OID oid;

    // If there was an error in the schema, return.
    if (error != 0) {
        *(Cor_exEnv->err_strm) << "WARNING :: error in schema\n";
        INVALIDATE_FID(fid);
        num_indexes = 0;
        return;
    }

    // Check if the relation is already defined. If so, return.
    if (SM.lookupFile(Cor_CArg(arg->functor()), arg->arity(), &oid)) {
        error = COR_DUP_SCHEMA;
        INVALIDATE_FID(fid);
        num_indexes = 0;
        return;
    }

    // If not, create a new file to hold the relation
    int e = sm_CreateFile(SM.bufGroup(),
        (temporary ? SM.tempvolid() : SM.volid()),
        &fid);
    if (ErrorCheck(e, "sm_CreateFile")) return;

    num_indexes = 0;
    count = 0;
    duplicates = !Cor_exEnv->C_check_subsum_default;
}

// *************************************************************************
void Cor_FileDescStruct::copyDesc(Cor_FileDescStruct *fdstruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return;

    fid = fdstruct->fid;
    for (int i = 0; i < COR_MAX_INDEXES; i++)
        indexes[i] = fdstruct->indexes[i];
    num_indexes = fdstruct->num_indexes;
    schema = fdstruct->schema;
    count = fdstruct->count;
}

// *************************************************************************
// Returns offset where an index entry can be made

int Cor_FileDescStruct::canAddIndex() 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return -1;

    if (num_indexes == COR_MAX_INDEXES) return -1;

    for (int i=0; i < COR_MAX_INDEXES; i++)
        if (!(indexes[i].valid())) return i;

    // This should not be reached
    return (-1);
}

// *************************************************************************
// Returns offset of the next valid index after offset i
// The first time, this is called with i = -1

int Cor_FileDescStruct::nextIndex(int i)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return -1;

    for (i += 1; i < COR_MAX_INDEXES; i++)
        if (indexes[i].valid()) return i;
    return (-1);
}

// *************************************************************************
// Adds an index at position 'offset' in the index array. The attributes
// of the index are specified by buf. Returns offset of the index

int Cor_FileDescStruct::addIndex(int offset, Cor_BitVector *bv)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    int match_index;
    // first find exact match. if it exists, then return its index
    char attrs[COR_MAX_ARGLIST_SIZE];
    for (int i = 0; i < schema.arity; i++)
        attrs[i] = bv->test(i);
    if ((match_index = findBestMatch(attrs)) >= 0) return match_index;

    // else create a new index
    if (!(indexes[offset].valid())) {
        indexes[offset].create(bv, schema.arity, 0);
        num_indexes++;
    }
    return offset;
}

// *************************************************************************
// Delete an index whose offset is specified 

void Cor_FileDescStruct::destroyIndex(int index_offset) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return;

    if (indexes[index_offset].valid()) {
        indexes[index_offset].destroy();
        num_indexes--;
    }
}

// *************************************************************************
// Adds an object to a specified index

void Cor_FileDescStruct::addObjectToIndex(int index_offset, USERDESC *u_desc) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return;

    if (indexes[index_offset].valid())
        indexes[index_offset].insert(u_desc, schema);
}

// *************************************************************************
// Find an index that best matches the given attr string
// IMPROVE THIS FUNCTION LATER !!

int Cor_FileDescStruct::findBestMatch(char *attr_str) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return -1;

    if (num_indexes == 0) return -1;

    for (int i = 0; i < COR_MAX_INDEXES; i++)
        if (indexes[i].valid() && (indexes[i].matches(attr_str, schema.arity)))
            return i;
    
    return -1;
}

// *************************************************************************
void Cor_FileDescStruct::print_index_adornments(ostream *outf)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized || !num_indexes) return;
    for (int i = 0; i < COR_MAX_INDEXES; i++) {
        if (indexes[i].valid()) {
            indexes[i].Cor_print_adornment(outf);
            *outf << " ";
        }
    }
}

// *************************************************************************

#endif WITH_PERSISTENCE
