// **************************************************************************
//    CoralStream.C
//
//    The CoralStream Class - A class to handle all I/O within Coral
//
// **************************************************************************

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <sys/types.h>

#ifdef    DO_PVM
#    include "PvmStream.h"
#else
#    include "SockStream.h"
#endif

#include "CoralStream.h"
#include "Debug.h"

// *************************************************************************
CoralStream::CoralStream(streambuf *tb, int delete_buffer)
: iostream(tb)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    orig = OST_NONE;
    _del_buffer = delete_buffer;
}

// *************************************************************************
CoralStream::~CoralStream()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    
    switch (orig) {

    default:
    case OST_NONE:
        if (_del_buffer)
            delete (this->rdbuf());
        break;

    case OST_IFS:
        delete ifs;
        break;

    case OST_OFS:
        delete ofs;
        break;
    }
}

// *************************************************************************
int   CoralStream::printf(char *fmt, ...)
#ifndef    NO_VFORM

{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int ret;

    assert(this);

    va_list ap;
    va_start(ap,fmt);

    ret = rdbuf()->vform(fmt, ap);

    va_end(ap);
    return(ret);
}

#else

{
// Always allocate a little extra, just in case.
#define  WIDTH_SLACK    8

// Need to allocate memory for anything larger than this.
#define  BUF_SZE      256

// If WIDTH_SLACK is too large, space will always need
// to be allocated.
#if (WIDTH_SLACK >= BUF_SZE)
#error "WIDTH_SLACK is too large."
#endif

#define PRINT_IT(ARG)     do {                            \
        switch (nargs)                                    \
        {                                                 \
            default:                                      \
            case 1:                                       \
                sprintf(tmpbuf, start, ARG);              \
                break;                                    \
                                                          \
            case 2:                                       \
                sprintf(tmpbuf, start,                    \
                    ((fld_width >= 0) ? fld_width         \
                                      : fld_prec), ARG);  \
                break;                                    \
                                                          \
            case 3:                                       \
                sprintf(tmpbuf, start, fld_width, fld_prec, ARG); \
                break;                                    \
        } } while(0)

    va_list         ap;
    int             j, len;

    char   buf[BUF_SZE];
    char  *fmt, *start, *end, *tmpbuf;
    char   sv;
    int    nargs, fld_width, fld_prec;
    long   width;
    int    is_short, is_long, is_longdoub;

    // The possible types we might see.
    char           *str;
    long            l;
    int             i;
    short           s;
    void           *v;
    double          d;
    long double     ld;

    va_start(ap,cfmt);

    fmt = (char*) cfmt;
    len = strlen(fmt);

    for (j=0; j < len; j++)
    {
      if (*fmt == '\0')
        break;
      else if (*fmt == '%')
      {
        start = fmt++;
        end   = fmt;

        // Handle the easy cases without using sprintf()
        switch (*fmt)
        {
            case 's':
                str = va_arg(ap, char*);
                *this << str;
                fmt++;
                break;

            case 'd':
            case 'i':
                i = va_arg(ap, int);
                *this << i;
                fmt++;
                break;
            
            case 'p':
                v = va_arg(ap, void*);
                *this << v;
                fmt++;
                break;

            case '%':
                *this << '%';
                fmt++;
                break;
            
            default:
                width = WIDTH_SLACK;   // width required to print it.
                fld_width = fld_prec = -1;
                nargs = 1;        // number of arguments to use.
                is_short = is_long = is_longdoub = 0;

                // Skip flags
                end += strspn(fmt, "-+ 0#");

                if (isdigit(*end))
                    width += strtol(end, &end, 10);
                else if (*end == '*')
                {
                    nargs++;
                    fld_width = va_arg(ap, int); 
                    end++;
                }

                if (*end == '.')
                    end++;

                if (isdigit(*end))
                    width += strtol(end, &end, 10);
                else if (*end == '*')
                {
                    nargs++;
                    fld_prec = va_arg(ap, int); 
                    end++;
                }

                if (*end == 'h')
                {
                    is_short = 1;
                    end++;
                }
                else if (*end == 'l')
                {
                    is_long = 1;
                    end++;
                }
                else if (*end == 'L')
                {
                    is_longdoub = 1;
                    end++;
                }

                if ((*end == 'h') || (*end == 'l') || (*end == 'L'))
                    end++;

                width += (end - start);

                if (width >= BUF_SZE)
                    tmpbuf = new char[width];
                else
                    tmpbuf = buf;

                // The next char should be the conversion char
                switch (*end)
                {
                    case 'd':
                    case 'i':
                    case 'o':
                    case 'x':
                    case 'X':
                    case 'u':
                        sv = *(++end);
                        *end = '\0';

                        if (is_short)
                        {
                            s = va_arg(ap, short);
                            PRINT_IT(s);
                        }
                        else if (is_long)
                        {
                            l = va_arg(ap, long);
                            PRINT_IT(l);
                        }
                        else
                        {
                            i = va_arg(ap, int);
                            PRINT_IT(i);
                        }

                        *end = sv;
                        *this << tmpbuf;
                        break;

                    case 'f':
                    case 'e':
                    case 'E':
                    case 'g':
                    case 'G':
                        sv = *(++end);
                        *end = '\0';

                        if (is_longdoub)
                        {
                            ld = va_arg(ap, long double);
                            PRINT_IT(ld);
                        }
                        else
                        {
                            d = va_arg(ap, double);
                            PRINT_IT(d);
                        }

                        *end = sv;
                        *this << tmpbuf;
                        break;

                    case 's':
                        sv = *(++end);
                        *end = '\0';
                        str = va_arg(ap, char*);
                        PRINT_IT(str);
                        *end = sv;
                        *this << tmpbuf;
                        break;

                    case 'p':
                        sv = *(++end);
                        *end = '\0';
                        v = va_arg(ap, void*);
                        PRINT_IT(v);
                        *end = sv;
                        *this << tmpbuf;
                        break;

                    case '%':
                        sv = *(++end);
                        *end = '\0';
                        // No argument required, 0 is a dummy one.
                        PRINT_IT(0);
                        *end = sv;
                        *this << tmpbuf;
                        break;

                    default:
                        *this << "<bad conversion char in printf>";
                        break;
                }

                if (width >= BUF_SZE)
                    delete [] tmpbuf;

                fmt = end;
                break;
        }
      }
      else
      {
        *this << *fmt;
        fmt++;
      }
    }
 
    va_end(ap);

#undef BUF_SZE
#undef PRINT_IT
}

#endif    /* !NO_VFORM */

// *************************************************************************
int   CoralStream::puts(char *buff, int add_nl)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    assert(this);
    *this << buff;
    if (add_nl)
        *this << endl;

    if (this->good())
        return(1);
    else
        return(EOF);
}

// *************************************************************************
int   CoralStream::putch(int ch)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    assert(this);
    char c = ch;
    *this << c;

    if (this->good())
        return(1);
    else
        return(EOF);
}

// *************************************************************************
//   Simulate what we need, neither getline() or get() quite do it.
int CoralStream::gets(char *buff, int n)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    assert(this);
    
    char c;
    int  tot = 0;

    n--;

    do {
        this->get(c);

        if (this->eof()) {
            break;
        }

       *buff++ = c;
        tot++;

    } while ((tot < n) && (c != '\n'));

   *buff = '\0';

    return(tot);
}

// *************************************************************************
//#define    SERVER            1

// *************************************************************************
#ifdef TEST_STREAM

int main()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //unsigned short port = 5432;
    unsigned short port = 7;    // TCP/IP echo port
    char result[200];

#    ifdef    SERVER

#    else

        // CLIENT

    Cor_sockbuf *echobuf = new Cor_sockbuf("quarg.cs.wisc.edu", port);

    CoralStream *in_cs = new CoralStream(cin.rdbuf());
    CoralStream *out_cs = new CoralStream(cout.rdbuf());
    CoralStream *echo_cs = new CoralStream(echobuf);

    int j=2;

        // Output possibilites (all should work)
    // *(out_cs) << "hello" << endl;
    out_cs->printf("hello%d\n", j);
    //out_cs->puts("hello");
    *(echo_cs) << "From socket" << endl;

    *out_cs << endl << flush;
    *echo_cs << flush;

        // Input possibilites (all should work)
    // *(in_cs) >> result;
    //in_cs->scanf("%s", result);
    //in_cs->gets(result,4);

    echo_cs->gets(result,20);

    cout << "Result is: " << result << endl;

#    endif  

    delete in_cs;
    delete out_cs;

    return(0);
}
#endif

// *************************************************************************
