// *************************************************************************
//    RuleFactor.C
// *************************************************************************

#include <stdio.h>

#include "RuleFactor.h"

#include "FactorPredList.h"
#include "FullMagic.h"
#include "SupMagic.h"
#include "Linearity.h"
#include "ExportInfo.h"
#include "Globals.h"
#include "Unparse.h"

// *************************************************************************

static int in_factor_list(FactorPredList &factor_preds, Cor_Literal *lit);
static void add_literal(FactorPredList &factor_preds, Cor_Literal *lit);
static int add_variable(Cor_Clause *rule, Cor_CArg& name);
static Cor_CArg *make_new_arg(Cor_Clause *rule);
static Cor_Literal *make_new_lit(Cor_Literal *lit, int arity, pred_kind kind);
static void new_bound_args(Cor_Clause *rule, Cor_Literal *lit);
static void new_leftmost_args(Cor_Clause *rule, Cor_Literal *lit);
static void free_args_from_free(Cor_Literal *lit1, Cor_Literal *lit2);
static void bound_args_from_bound(Cor_Literal *lit1, Cor_Literal *lit2);

#ifdef    NOT_USED
static void copy_all_args(Cor_Literal *lit1, Cor_Literal *lit2);
static void bound_args_from_leftmost(Cor_Literal *lit1, Cor_Literal *lit2);
#endif

static void rightmost_args_from_bound(Cor_Literal *lit1, Cor_Literal *lit2);
static void leftmost_args_from_bound(Cor_Literal *lit1, Cor_Literal *lit2);
static void leftmost_args_from_leftmost(Cor_Literal *lit1, Cor_Literal *lit2);
static Cor_Literal *make_ord_lit(Cor_Literal *oldhead);
static Cor_Literal *make_magic_factor_lit(Cor_Literal *lit );
static void make_derived_magic_rules(Cor_Clause *rule,
    Cor_Collection *magic_rules, int pred_index);
static void do_factor_non_recursive(Cor_Clause *rule,
    Cor_Collection *magic_rules);
static void do_factor_rt_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules);
static void do_factor_lt_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules);
static void do_factor_pseudo_lt_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules);
static void do_factor_multi_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules);

// *************************************************************************

static FactorPredList factor_linear_preds, factor_magic_preds;

// *************************************************************************
static int in_factor_list(FactorPredList &factor_preds, Cor_Literal *lit)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    struct FactorPredList *temp = factor_preds.next;

    while (temp != NULL) {
        if ((temp->lit->pred.equals(lit->pred)) &&
            (temp->lit->adorn == lit->adorn) &&
            (temp->lit->args.count() == lit->args.count())) return 1;
        temp = temp->next;
    }
    return 0;
}

// *************************************************************************
static void add_literal(FactorPredList &factor_preds, Cor_Literal *lit)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    struct FactorPredList *temp;

    if (in_factor_list(factor_preds, lit)) return;

    temp = new FactorPredList;
    temp->lit = lit;
    temp->next = factor_preds.next;
    factor_preds.next = temp;
}

// *************************************************************************
static int add_variable(Cor_Clause *rule, Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *carg = new Cor_CArg(name);

    if (! rule->var_names)
        rule->var_names = new Cor_CArgStack();

    rule->var_names->push(carg);
    return(rule->var_names->count() - 1);
}

// *************************************************************************
static Cor_CArg *make_new_arg(Cor_Clause *rule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    static int count = 0;
    char buf[10];

    sprintf(buf, "C__F%d", count++);
    Cor_CArg *ret = new Cor_CArg(buf);
    ret->set_var(add_variable(rule, *ret));
    return(ret);
}

// *************************************************************************
static Cor_Literal *make_new_lit(Cor_Literal *lit, int arity, pred_kind kind)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *new_lit; 

    new_lit = (Cor_Literal*) new Cor_Literal(arity);
    new_lit->kind = kind;
    new_lit->base_pred = ((kind == COR_MAGIC) ? 0 : lit->base_pred);
    new_lit->negated = ((kind == COR_MAGIC) ? 0 : lit->negated);
    new_lit->pred = lit->pred;

    return new_lit;
}

// *************************************************************************
//    Replaces bound args of lit by newly created variables.

static void new_bound_args(Cor_Clause *rule, Cor_Literal *lit)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit != NULL && rule != NULL);
    for (int i=0; i< lit->args.count(); i++) {
        if (lit->adorn[i]) {
            // The BOUND variables are replaced by new variables.
            lit->args[i] = *(make_new_arg(rule));
        }
    }
}

// *************************************************************************
//    Replaces left 1/2 of args of lit by newly created variables.
//    Used only for factor-magic literals.

static void new_leftmost_args(Cor_Clause *rule, Cor_Literal *lit)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit != NULL && rule != NULL);
    int len_by_2 = lit->args.count()/2;

    for (int i=0; i < len_by_2; i++)
        lit->args[i] = *(make_new_arg(rule));
}

// *************************************************************************
//    Replaces free args of lit1 by free args of lit2.  The free arg counts of
//    the two should be equal.  Called with two literals of the same predicate.

static void free_args_from_free(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit1 != NULL && lit2 != NULL);
    int i = 0, j = 0;

    for (; i< lit1->args.count(); i++) {
        if (!lit1->adorn[i]) {
            while ((j < lit2->args.count()) && (lit2->adorn[j])) j++;

            //    The FREE variables in lit1 are replaced by FREE variables
            //    from lit2.
            lit1->args[i] = lit2->args[j++];
        }
    }
}

// *************************************************************************
//    Replaces bound args of lit1 by bound args of lit2.  The bound arg counts
//    of the two should be the same.  Called with two literals of the same
//    predicate.

static void bound_args_from_bound(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit1 != NULL && lit2 != NULL);
    int i = 0, j = 0;

    for (; i< lit1->args.count(); i++) {
        if (lit1->adorn[i]) {
            while ((j < lit2->args.count()) && (!lit2->adorn[j])) j++;

            //    The BOUND variables in lit1 are replaced by 
            //    bound variables from lit2.
            lit1->args[i] = lit2->args[j++];
        }
    }
}

#ifdef    NOT_USED
// *************************************************************************
// Replaces all args of lit1 by args of lit2.  the arglist lengths shoud be
// the same.  Called with two literals of the same predicate.

static void copy_all_args(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit1 != NULL && lit2 != NULL);
    int i = 0;

    for (; i< lit1->args.count(); i++) {
        lit1->args[i] = lit2->args[i];
    }
}

#endif
// *************************************************************************
// Replaces rightmost 1/2 of the args of lit1 by bound args of lit2.  Called
// with a factor-magic literal and a literal of the same predicate.  lit1's
// arglist is twice the length of the bound args of lit2.
static void rightmost_args_from_bound(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit1 != NULL && lit2 != NULL);
    int i, j = 0;
    for (i = lit1->args.count()/2; i< lit1->args.count(); i++) {
        while (!lit2->adorn[j]) j++;
        ASSERT(j < lit2->args.count());
        lit1->args[i] = lit2->args[j++];
    }
}

// *************************************************************************
// Replaces leftmost 1/2 of the args of lit1 by bound args of lit2.  Called
// with a factor-magic literal and a literal of the same predicate.  lit1's
// arglist is twice the length of the bound args of lit2.

static void leftmost_args_from_bound(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit1 != NULL && lit2 != NULL);
    int i = 0, j = 0;
    for (; i< lit1->args.count()/2; i++) {
        while (!lit2->adorn[j]) j++;
        ASSERT(j < lit2->args.count());
        //    The leftmost variables in lit1 are replaced by 
        //    bound variables from lit2.
        lit1->args[i] = lit2->args[j++];
    }
}

// *************************************************************************
// Replaces leftmost 1/2 of the args of lit1 by the leftmost 1/2 of the args
// of lit2.  Called with two factor-magic literals of the same predicate.

static void leftmost_args_from_leftmost(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit1 != NULL && lit2 != NULL);
    int i = 0;
    for (; i< lit1->args.count()/2; i++)
        lit1->args[i] = lit2->args[i];
}

#ifdef    NOT_USED
// *************************************************************************
// Replaces bound args of lit1 by the leftmost 1/2 of the args of lit2.
// Called with a literal and a factor-magic literal of the same predicate.
// lit2's arglist length is twice the bound args of lit1.
static void bound_args_from_leftmost(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(lit1 != NULL && lit2 != NULL);
    int i = 0;
    int j = 0;

    for (; i< lit1->args.count()/2; i++) {
        while (!lit1->adorn[j]) j++;
        ASSERT(j < lit1->args.count());

        lit1->args[j++] = lit2->args[i];
    }
}

#endif
// *************************************************************************
static Cor_Literal *make_ord_lit(Cor_Literal *oldhead)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *new_lit; 

    new_lit = make_new_lit(oldhead, oldhead->arity(), COR_REGULAR);
    new_lit->adorn = oldhead->adorn;

    return new_lit;
}

// *************************************************************************
static Cor_Literal *make_magic_factor_lit(Cor_Literal *lit )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *new_lit;

    int count = lit->adorn.count(); // count of bound args in lit
    new_lit = make_new_lit(lit, 2*count, COR_FACTORMAGIC);
    new_lit->adorn = lit->adorn;

    //    Add to list of factor/magic literals
    add_literal(factor_magic_preds, new_lit);

    return new_lit;
}

// *************************************************************************
// Looks at all the factor-magic predicates newly created, and creates rules
// of the form mc_p(C,C) :- m_p(C).  for each of them.

void add_default_rules(Cor_Collection *magicrules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *rule, *factorrule;
    int i, j, nrules;
    struct FactorPredList *temp = factor_magic_preds.next;

    int num_org_rules = 1;

    nrules = magicrules->count();

    for (j=0; j < nrules; j++) {
      factorrule = magicrules->get_ith(j);

      if (!(factorrule->head->kind == COR_MAGIC || 
                factorrule->head->kind == COR_SUPMAGIC))
        num_org_rules++;
    }

    while (temp != NULL) {
        int count = temp->lit->adorn.count();
        rule = Cor_copy_rule(NULL);
        rule->head = make_new_lit(temp->lit, count*2, COR_FACTORMAGIC);
        rule->head->adorn = temp->lit->adorn;

        if (temp->lit->adorn.count() == 0) {
            delete rule->var_names;
            rule->var_names = NULL;
        } else {
            delete rule->var_names;
            rule->var_names = new Cor_CArgStack();

            int j = -1;
            int lim = temp->lit->adorn.count();
            for (i=0; i < lim; ++i) {
                Cor_CArg *carg = new
                        Cor_CArg(temp->lit->args[j=temp->lit->adorn.next(j)]);
                rule->var_names->push(carg);
            }
        }

        for (i=0; i< count; i++)
            rule->head->args[i] = *(make_new_arg(rule));
        for (i = count; i < count*2; i++)
            rule->head->args[i] = rule->head->args[i-count];

        Cor_alloc_preds(rule, 1);

        rule->preds[0] = make_new_lit(rule->head, count, COR_MAGIC);
        rule->preds[0]->adorn = rule->head->adorn;

        for (i=0; i< rule->preds[0]->args.count(); i++)
            rule->preds[0]->args[i] = rule->head->args[i];

            // Give a 'fake' label to this rule
        if (rule->clause_names->is_empty()) {
            rule->clause_names->add(num_org_rules++,0);
        }

        magicrules->push(rule);

        temp = temp->next;
    }
}

// *************************************************************************
// Similar to the fn create_derived_magic_rules in FullMagic.C in the magic
// directory.  Creates factor_magic rules for derived predicates on the rhs
// of a factored rule.  Uses standard l-to-r sips to generate rules.  Does
// not use any user-provided sips, because their meaning in the factoring
// context is unclear.  Have I over-simplified this ??? -- PRAVEEN

static void make_derived_magic_rules(Cor_Clause *rule,
    Cor_Collection *magic_rules, int pred_index)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *magicrule;

    if ((rule->preds[pred_index]->base_pred) ||
        (rule->preds[pred_index]->kind == COR_MAGIC)) return;

    //    Copy proto header
    magicrule = Cor_copy_rule(rule);
    magicrule->head = build_magic_pred(rule, rule->preds[pred_index], 0);
    Cor_alloc_preds(magicrule, pred_index);
    for (int i = 0; i< pred_index; i++)
        magicrule->preds[i] =  rule->preds[i];

    magicrule->type = COR_MAGICRULE;
    magicrule->source_rule = rule;
    magicrule->source_pred = rule->preds[pred_index];
    magic_rules->push(magicrule);
}

// *************************************************************************
// Creates factored/magic rules from the rules in 'factor_rules'.  The
// factor_status of these should already have been assigned.

void factor_rewrite(Cor_Collection *factor_rules,
    Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *single_rule;
    int j, nrules;

    factor_linear_preds.init();
    factor_magic_preds.init();

    nrules = factor_rules->count();

    for (j=0; j < nrules; j++) {
        single_rule = factor_rules->get_ith(j);

        switch (single_rule->factor_status) {
        case COR_RT_LINEAR:
            do_factor_rt_linear(single_rule, magic_rules);
            add_literal(factor_linear_preds, single_rule->head);
            break;
        case COR_LT_LINEAR:
            // If UseFactorMagic is set, lt_linear rules are treated
            // the same as pseudo_lt_linear rules
            if (Cor_parserStruct->CurModule->UseFactorMagic)
                do_factor_pseudo_lt_linear(single_rule, magic_rules);
            else
                do_factor_lt_linear(single_rule, magic_rules);
            add_literal(factor_linear_preds, single_rule->head);
            break;
        case PSEUDO_LT_LINEAR:
            do_factor_pseudo_lt_linear(single_rule, magic_rules);
            add_literal(factor_linear_preds, single_rule->head);
            break;
        case COR_MULTI_LINEAR:
            do_factor_multi_linear(single_rule, magic_rules);
            add_literal(factor_linear_preds, single_rule->head);
            break;
        case COR_NON_LINEAR:
            if (Cor_parserStruct->CurModule->UseSupplementaryMagic)
                do_supp_magic_on_rule(single_rule, magic_rules);
            else
                do_magic_on_rule(single_rule, magic_rules);
            break;    
        }
    }

    for (j=0; j < nrules; j++) {
        single_rule = factor_rules->get_ith(j);

        if (single_rule->factor_status == COR_NON_RECURSIVE) {
            if (in_factor_list(factor_linear_preds, single_rule->head)) {
                do_factor_non_recursive(single_rule, magic_rules);
            } else if (Cor_parserStruct->CurModule->UseSupplementaryMagic) {
                do_supp_magic_on_rule(single_rule, magic_rules);
            } else {
                do_magic_on_rule(single_rule, magic_rules);
            }
        }
    }

    add_default_rules(magic_rules);
}

// *************************************************************************
// Rewrite non-recursive rules, according to Context Rule Factoring.

static void do_factor_non_recursive(Cor_Clause *rule,
    Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *magicrule;

    magicrule = Cor_copy_rule(rule);
    Cor_alloc_preds(magicrule, rule->num_preds + 1);

    magicrule->head = make_ord_lit( rule->head);
    magicrule->head_deletes_count = rule->head_deletes_count;
    magicrule->head_deletes = rule->head_deletes;
    new_bound_args(magicrule, magicrule->head);
    free_args_from_free(magicrule->head, rule->head);

    magicrule->preds[0] = make_magic_factor_lit( rule->head);
    leftmost_args_from_bound(magicrule->preds[0], magicrule->head);
    rightmost_args_from_bound(magicrule->preds[0], rule->head);

    int j,i;
    for (j=1,i=0;  i<rule->num_preds; ++j, ++i)
        magicrule->preds[j] = rule->preds[i];

    magic_rules->push(magicrule);
    for (i=1; i < magicrule->num_preds; ++i)
        make_derived_magic_rules(magicrule, magic_rules, i);
}

// *************************************************************************
// Rewrite right linear rules, according to Context Rule Factoring.

static void do_factor_rt_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *magicrule;

    magicrule = Cor_copy_rule(rule);
    Cor_alloc_preds(magicrule, rule->num_preds);

    magicrule->head = make_magic_factor_lit( rule->head);

    new_leftmost_args(magicrule, magicrule->head);
    rightmost_args_from_bound(magicrule->head, rule->factor_pred);

    magicrule->preds[0] = make_magic_factor_lit( rule->head);
    leftmost_args_from_leftmost(magicrule->preds[0], magicrule->head);
    rightmost_args_from_bound(magicrule->preds[0], rule->head);

    int j,i;
    for (j=1,i=0;    i<rule->num_preds; ++i) {
        if (rule->preds[i] == rule->factor_pred) continue;
        magicrule->preds[j] = rule->preds[i];
        j++;
    }

    magic_rules->push(magicrule);
    for (i=1; i < magicrule->num_preds; ++i)
        make_derived_magic_rules(magicrule, magic_rules, i);
}

// *************************************************************************
// Rewrite left linear rules, according to Context Rule Factoring.

static void do_factor_lt_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *magicrule;

    magicrule = Cor_copy_rule(rule);
    Cor_alloc_preds(magicrule, rule->num_preds);

    magicrule->head = make_ord_lit( rule->head);
    magicrule->head_deletes_count = rule->head_deletes_count;
    magicrule->head_deletes = rule->head_deletes;
    new_bound_args(magicrule, magicrule->head);
    free_args_from_free(magicrule->head, rule->head);

    magicrule->preds[0] = make_ord_lit( rule->head);
    bound_args_from_bound(magicrule->preds[0], magicrule->head);
    free_args_from_free(magicrule->preds[0], rule->factor_pred);

    int j,i;
    for (j=1,i=0;    i<rule->num_preds; ++i) {
        if (rule->preds[i] == rule->factor_pred) continue;
        magicrule->preds[j] = rule->preds[i];
        j++;
    }

    magic_rules->push(magicrule);
    for (i=1; i < magicrule->num_preds; ++i)
        make_derived_magic_rules(magicrule, magic_rules, i);
}

// *************************************************************************
// Rewrite pseudo left linear rules, according to Context Rule Factoring.

static void do_factor_pseudo_lt_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *magicrule;

    magicrule = Cor_copy_rule(rule);
    Cor_alloc_preds(magicrule, rule->num_preds+1);

    magicrule->head = make_ord_lit( rule->head);
    magicrule->head_deletes_count = rule->head_deletes_count;
    magicrule->head_deletes = rule->head_deletes;
    new_bound_args(magicrule, magicrule->head);
    free_args_from_free(magicrule->head, rule->head);

    magicrule->preds[0] = make_magic_factor_lit( rule->head);
    leftmost_args_from_bound(magicrule->preds[0], magicrule->head);
    rightmost_args_from_bound(magicrule->preds[0], rule->head);

    magicrule->preds[1] = make_ord_lit( rule->head);
    bound_args_from_bound(magicrule->preds[1], magicrule->head);
    free_args_from_free(magicrule->preds[1], rule->factor_pred);

    int j,i;
    for (j=2,i=0;    i<rule->num_preds; ++i) {
        if (rule->preds[i] == rule->factor_pred) continue;
        magicrule->preds[j] = rule->preds[i];
        j++;
    }

    magic_rules->push(magicrule);
    for (i=1; i < magicrule->num_preds; ++i)
        make_derived_magic_rules(magicrule, magic_rules, i);
}

// *************************************************************************
// Rewrite multi linear rules, according to Context Rule Factoring.

static void do_factor_multi_linear(Cor_Clause *rule,
    Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *magicrule;
    int num_preds, j_start = 0;

    magicrule = Cor_copy_rule(rule);

    num_preds = (Cor_parserStruct->CurModule->UseFactorMagic ?
        rule->num_preds : rule->num_preds-1);
    Cor_alloc_preds(magicrule, num_preds);

    magicrule->head = make_magic_factor_lit( rule->head);

    new_leftmost_args(magicrule, magicrule->head);
    rightmost_args_from_bound(magicrule->head, rule->factor_pred);

    if (Cor_parserStruct->CurModule->UseFactorMagic) {
        magicrule->preds[0] = make_magic_factor_lit( rule->head);
        leftmost_args_from_leftmost(magicrule->preds[0], magicrule->head);
        rightmost_args_from_bound(magicrule->preds[0], rule->head);

        j_start = 1;
    }

    int j,i;
    for (j=j_start,i=0;    i<rule->num_preds; ++i) {
        if (rule->preds[i] == rule->factor_pred) continue;
        if (rule->head->predicate != rule->preds[i]->predicate) {
            magicrule->preds[j] = rule->preds[i];
            j++;
            continue;
        }
        magicrule->preds[j] = make_ord_lit( rule->preds[i]);
        bound_args_from_bound(magicrule->preds[j], magicrule->head);
        free_args_from_free(magicrule->preds[j], rule->preds[i]);
        j++;
    }

    magic_rules->push(magicrule);
    for (i=1; i < magicrule->num_preds; ++i)
        make_derived_magic_rules(magicrule, magic_rules, i);
}

// *************************************************************************
